--[[============================================================================
main.lua
============================================================================]]--

-- Placeholder for the dialog
--local dialog = nil

-- Placeholder to expose the ViewBuilder outside the show_dialog() function
--local vb = nil

local rs

-- Reload the script whenever this file is saved. 
-- Additionally, execute the attached function.
_AUTO_RELOAD_DEBUG = function()
  
end

-- Read from the manifest.xml file.
class "RenoiseScriptingTool" (renoise.Document.DocumentNode)
  function RenoiseScriptingTool:__init()    
    renoise.Document.DocumentNode.__init(self) 
    self:add_property("Name", "Untitled Tool")
    self:add_property("Id", "Unknown Id")
  end

local manifest = RenoiseScriptingTool()
local ok,err = manifest:load_from("manifest.xml")
local tool_name = manifest:property("Name").value
local tool_id = manifest:property("Id").value





---------------------------------------  
function extend_group_left(group_index)  
---------------------------------------
  
  local song = renoise.song()
  local group_track = song:track(group_index)

  --get the index of the track to the left of the group.  This is the track we want to add to the group
  local left_of_grp_idx = (group_index - #group_track.members - 1) 
  
  --if left_of_grp_idx == 0 then there are no more tracks to the left so return early 
  if left_of_grp_idx == 0 then 
    return
  end
  
  --if group has no members already, then it is a simple add track to group operation and
  --no re-organisation is needed
  if #group_track.members == 0 then
    song:add_track_to_group(left_of_grp_idx,group_index)
    return
  end
  
  --1) Add 2 dummy tracks into group
  ----------------------------------
  --NOTE We need to insert 2 dummy tracks to account for bugs with the renoise swap tracks behaviour within groups
  --see:  http://forum.renoise.com/index.php/topic/51960-lua-api-bug-swap-tracks/
  
  --set member_idx to the first member track of the group
  local member_idx = left_of_grp_idx + 1

  --loop group members to find insert point for dummy tracks (after first legitimate member; i.e. not a track in a further sub-group)
  for i = 1,#group_track.members do
    --check if member is not in a sub group by checking its parent
    if rawequal(song:track(member_idx).group_parent,group_track) == true then
      break
    end
    --if here then member was in a sub- group so increment member_idx until we get to a legitimate track
    member_idx = member_idx + 1
  end
  
  --again due to insert_track behaviour we need to insert the dummy tracks after the first present and legitimate member of the group
  --so + 1
  local insert_point = member_idx + 1
  
  --insert 2 dummy tracks at member_idx 
  local dummy_trk_2 = song:insert_track_at(insert_point)
  local dummy_trk_1 = song:insert_track_at(insert_point)
  
  --name the tracks for bug hunting i.e. if they are not deleted we will see them in pattern ed
  dummy_trk_1.name = "Dummy Track 1"
  dummy_trk_2.name = "Dummy Track 2"
  
  --2) Swap Dummy tracks to members 1 and 2 of group (this works re. the swap bug as we are not swapping two groups)
  song:swap_tracks_at(member_idx,member_idx+1)
  song:swap_tracks_at(member_idx+1,member_idx+2)
  
  --3) Swap dummy member 2 with target track outside group (left_of_grp_idx was found near beginning of function)
  --note: if we were to swap dummy member 1 then both tracks get ejected from group
  song:swap_tracks_at(left_of_grp_idx,(left_of_grp_idx + 2))
 
  --4) delete dummy tracks by looping objects 
  --(as we are deleting tracks it is safer to loop and find the objects specifically or
  -- we might lose wanted song tracks if there was an arithmetic mistake if a bug is present)
  for i = 1,#song.tracks do
    if rawequal(song:track(i),dummy_trk_1) then
      --delete dummy 1
      renoise.song():delete_track_at(i)
      break
    end
  end
  
  for i = 1,#song.tracks do
    if rawequal(song:track(i),dummy_trk_2) then
    --delete dummy 2
      renoise.song():delete_track_at(i)
      break
    end
  end
end 




--------------------------------------------------------------------------------
-- Main functions
--------------------------------------------------------------------------------
local function group_selected_tracks()
 
  local song = renoise.song()
  local start_track = nil 
  local end_track = nil
  local selected_track_holder = song.selected_track_index
  
  --if no selection then set the selection to selected track
  if song.selection_in_pattern == nil then
    start_track = song.selected_track_index
    end_track = song.selected_track_index
  else
    --set start and end_track from selection
    start_track = song.selection_in_pattern.start_track
    end_track = song.selection_in_pattern.end_track
  end
    
  --if master or sends are selected, then set the end_track to the last sequencer track in renoise
  if end_track > song.sequencer_track_count then
    end_track = song.sequencer_track_count
  end
  
 
  --loop left to find leftmost track that will be included in the new group
  -------------------------------------------------------------------------
  --we need to take account of subgroups here
  local left_most_track_index = start_track
  --loop
  for i = end_track,start_track, -1 do
    local trk = song:track(i)
    --sub groups can extend further left than the pattern selection and they will still be added
    --so we find out their left most track member index here
    if trk.type == renoise.Track.TRACK_TYPE_GROUP then
      if (i - #trk.members) < left_most_track_index then
        left_most_track_index = (i - #trk.members)
      end
    end
  end
  
  ------------------------------------------
  --due to renoise only allowing nesting to 6 levels (5 parent groups).  We need to check
  --if addingthe newly added group now violates the limit.  There is due to a current API bug where
  --you can bypass this limit without a scripting error being fired. See:
  --http://forum.renoise.com/index.php/topic/52010-lua-api-bug-nesting-groups-allowed-beyond-6-levels/
 
  
  --function that will return the amount of parent groups that any track has
  ----------------------------------------------------
  local function get_tracks_nesting_level(track_index)
  ----------------------------------------------------
  local song = renoise.song()
  --default to selected track index if none passed in
  track_index = track_index or song.selected_track_index
  --track we are testing
  local track = song:track(track_index) 
  local track_parents = 0 --counter
  local next_parent = track.group_parent --parent group object or nil
  
  --breaks when there are no further parent groups to count
  while (next_parent ~= nil) do
    --get next parent group object
    next_parent = next_parent.group_parent 
    --increment counter
    track_parents = track_parents + 1
  end
    return track_parents
  end
  ----------------------------------------------------

   
  --when a the new group is added there may be tracks further to the left of the selection added 
  --i.e. if left of selection was in a group itself that is part of the selection
  --new group member track range
  local last_new_member = end_track
  local first_new_member = left_most_track_index

  local total_parent_count = 0
  --loop all tracks that will be in the new group and get the number of parents 
  --for the most deeply nested track
  for i = first_new_member,last_new_member do
    local cur_par_count = get_tracks_nesting_level(i) --local function() above
    if cur_par_count > total_parent_count then
      total_parent_count = cur_par_count
    end
  end
  
  --if the new group violates the renoise limit of 5 parent groups then return
  if total_parent_count > 4 then
    --reset selected track to original  
    song.selected_track_index = selected_track_holder
    renoise.app():show_status("Maximum 6 Levels Of Nesting Reached")
    return
  end
  
  --new group will be valid so:
  --insert a group at end_track +1 position, passed the end track
  local new_group = song:insert_group_at(end_track + 1)
  
  --reverse loop to extend the group left using extend_group_left() function [defined outside of main()]
  for i = end_track, start_track, -1 do
    --extend the group left, bypassing any tracks that do not have the same parent as the newly inserted group track
    if rawequal(song:track(i).group_parent,new_group.group_parent) == true then
      extend_group_left(end_track + 1) 
    end
   --removed due to API related bugs
   --see: http://forum.renoise.com/index.php/topic/51960-lua-api-bug-swap-tracks/
   --song:add_track_to_group(st, et+1)
  end
  
  --reset selected track index
  song.selected_track_index = selected_track_holder
   
end

-- This example function is called from the GUI below.
-- It will return a random string. The GUI function displays 
-- that string in a dialog.
--local function get_greeting()
  --local words = {"Hello world!", "Nice to meet you :)", "Hi there!"}
  --local id = math.random(#words)
  --return words[id]
--end


--------------------------------------------------------------------------------
-- Menu entries
--------------------------------------------------------------------------------

--renoise.tool():add_menu_entry {
  --name = "Main Menu:Tools:Group Selected Tracks",
  --invoke = group_selected_tracks
--}


--------------------------------------------------------------------------------
-- Key Binding
--------------------------------------------------------------------------------

renoise.tool():add_keybinding {
  name = "Global:Tools:Group Selected Tracks",
  invoke = group_selected_tracks
}


--------------------------------------------------------------------------------
-- MIDI Mapping
--------------------------------------------------------------------------------

--[[
renoise.tool():add_midi_mapping {
  name = tool_id..":Show Dialog...",
  invoke = show_dialog
}
--]]
